import re
import uvicore
import inspect
from typing import Dict, List, Any, Union, Callable
from uvicore.support.dumper import dump, dd
from types import SimpleNamespace as obj
from collections import namedtuple
from uvicore.contracts import Dispatcher as DispatcherInterface
from uvicore.support import module


@uvicore.service('uvicore.events.dispatcher.Dispatcher',
    aliases=['Dispatcher', 'dispatcher', 'Event', 'event'],
    singleton=True,
)
class Dispatcher(DispatcherInterface):
    """Event Dispatcher private class.

    Do not import from this location.
    Use the uvicore.events singleton global instead."""

    @property
    def events(self) -> Dict[str, Dict]:
        return self._events

    @property
    def listeners(self) -> Dict[str, List]:
        return self._listeners

    @property
    def wildcards(self) -> List:
        return self._wildcards

    def __init__(self) -> None:
        self._events = {}
        self._listeners = {}
        self._wildcards = []

    def register(self, events: Dict[str, Dict]):
        # Ensure each event has a 'name' attribute
        for (key, value) in events.items():
            if 'name' not in value: value['name'] = key

        # Merge (but not deep) with existing events
        # Merging allows override by other later defined packages
        self._events = {**self.events, **events}

    def handle(self, events: Union[str, List], ):
        # Decorator to listen to events, same as uvicore.event.listen('someevent', handler)
        def decorator(func):
            self.listen(events, func)
            return func
        return decorator

    def listen(self, events: Union[str, List], listener: Union[str, Callable] = None) -> None:
        # Decorator usage
        if not listener: return self.handle(events)

        if type(events) == str: events = [events]

        # Do NOT check self.events for listener verification becuase self.events
        # is probably empty at this stage.  All registrations and listeners are
        # applied in service provider register() methods so cannot rely on
        # self.events being full and complete.  This means we cannot propert
        # match wildcard events to actual events.

        # Expand all wildcards
        for event in events:
            if event not in self.listeners:
                self._listeners[event] = []
            self._listeners[event].append(listener)
            if '*' in event:
                self._wildcards.append(event)

    def subscribe(self, listener: Union[str, Callable]) -> None:
        try:
            if type(listener) == str:
                module.load(listener).object().subscribe(uvicore.events)
            else:
                listener.subscribe(uvicore.events)
        except ModuleNotFoundError:
            pass

    def dispatch(self, event: Any, payload = {}) -> None:
        # When using this actual uvicore.events.dispatch(MyEvent())
        # In other words, passing in a class instance, we want to call
        # the class .dispatch() method itself instead if just passing it
        # on to _dispatch below.  Why?  Because users will override the
        # dispatch method and we always want to make sure that runs.
        # If users call MyEvent().dispatch() this method is never hit
        # and _dispatch is called directly.

        # Get event by string name or class inspection
        event_meta = self.event(event)

        # Event not registered, dispatch nothing
        # No, because we have tons of dynamic events, like the ORM
        #if not event_meta: return

        if type(event) == str:
            if '-' in event:
                self._dispatch(event, payload)
            else:
                # See if string event has a matching class.  If so, import and dispatch it
                found = module.location(event)
                if found:
                    module.load(event).object(**payload).dispatch()
                else:
                    # No class found for this string.  This is OK because events can
                    # be strings without matching classes.  Dispatch it anyway
                    self._dispatch(event, payload)
        else:
            # Event is a class.  Call the actual classes dispatch method
            # in case the user overrode it, we still execute it
            event.dispatch()

    def _dispatch(self, event: Any, payload = {}) -> None:
        # Get event by string name or class inspection
        event_meta = self.event(event)

        # Event not registered, dispatch nothing
        # No, because we have tons of dynamic events, like the ORM
        #if not event_meta: return

        if type(event) == str:
            # Event is a string.  Convert payload into an object
            # so the data is accessible as if it were a class based event
            #payload['event'] = event_meta
            if payload:
                payload = namedtuple('payload', sorted(payload))(**payload)
            else:
                payload = None
        else:
            # Event is a class or method.  The payload is the instantiated
            # class itself and the event is the classes MODULE name, not the class name itself
            payload = event
            event = str(event.__class__).split("'")[1]
            #event_meta = self.events.get(event)
            #event_meta['name'] = event
            #setattr(payload, 'event', event_meta)

        # Fire all event listeners
        self._fire_listeners(event_meta, payload)

    def _fire_listeners(self, event: Dict, payload: Dict) -> None:
        listeners = self.event_listeners(event['name'])
        for listener in listeners:
            # New up the listener
            if type(listener) == str:
                try:
                    module.load(listener).object(uvicore.app).handle(event, payload)
                except ModuleNotFoundError:
                    # Bad listener, skip
                    pass
            else:
                listener(event, payload)

    def event(self, event: Union[str, Callable]) -> Dict:
        if type(event) == str:
            name = event
        else:
            name = str(event.__class__).split("'")[1]
        if name in self.events:
            # dump('before', self.events.get(name))
            # event_meta = self.events.get(name)
            # event_meta['name'] = name
            # dump('after', self.events.get(name))
            #return event_meta
            return self.events.get(name)
        else:
            # This event is NOT registered, but we still want it to work
            # because of all the dynamic events like ORM makes
            # So create a fake event meta
            return {
                'name': name
            }

    def event_listeners(self, event: str) -> List:
        listeners = []
        if event in self.listeners:
            listeners += self.listeners[event]

        for wildcard in self.wildcards:
            regex = wildcard
            #regex = wildcard.replace('*', '.*')
            dump(regex)
            if re.search(regex, event):
                listeners += self.listeners[wildcard]

        return listeners


# IoC Class Instance
# **Not meant to be imported from here**.  Use the uvicore.events singleton global instead.
# Only here because uvicore bootstrap needs to import it without a service provider.
# By using the default bind and make feature of the IoC we can swap the implimentation
# at a high bootstrap level using our app configs 'bindings' dictionary.
# The only two classes that do this are Application and the event Dispatcher.
#Dispatcher: _Dispatcher = uvicore.ioc.make('Dispatcher', _Dispatcher, singleton=True, aliases=['dispatcher', 'Event', 'event', 'Events', 'events'])



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Union, Mapping, Optional

import sqlalchemy as sa
from databases import Database as EncodeDatabase
from sqlalchemy.sql import ClauseElement
from uvicore.contracts import DbQueryBuilder

from .connection import Connection
from .package import Package


class Database(ABC):
    pass

    @property
    @abstractmethod
    def default(self) -> str:
        """The default connection str for the main running app"""
        pass

    @property
    @abstractmethod
    def connections(self) -> Dict[str, Connection]:
        """All connections from all packages, keyed by connection str name"""
        pass

    @property
    @abstractmethod
    def engines(self) -> Dict[str, sa.engine.Engine]:
        """All engines for all unique (by metakey) connections, keyed by metakey"""
        pass

    @property
    @abstractmethod
    def databases(self) -> Dict[str, EncodeDatabase]:
        """All Encode Databases for all unique (by metakey) connections, keyed by metakey"""
        pass

    @property
    @abstractmethod
    def metadatas(self) -> Dict[str, sa.MetaData]:
        """All SQLAlchemy Metadata for all unique (by metakey) connections, keyed by metakey"""
        pass

    @abstractmethod
    def init(self, default: str, connections: List[Connection]) -> None:
        """Initialize the database system with a default connection str and List of all Connections from all packages"""
        pass

    @abstractmethod
    def packages(self, connection: str = None, metakey: str = None) -> List[Package]:
        """Get all packages with the metakey (direct or derived from connection str)."""
        pass

    @abstractmethod
    def metakey(self, connection: str = None, metakey: str = None) -> str:
        """Get one metekay by connection str or metakey"""
        pass

    @abstractmethod
    def connection(self, connection: str = None) -> Connection:
        """Get one connection by connection name"""
        pass

    @abstractmethod
    def metadata(self, connection: str = None, metakey: str = None) -> sa.MetaData:
        """Get one SQLAlchemy Metadata by connection str or metakey"""
        pass

    @abstractmethod
    def table(self, table: str, connection: str = None) -> sa.Table:
        """Get one SQLAlchemy Table by name (without prefix) and connection str or connection.tablename dot notation"""
        pass

    @abstractmethod
    def tablename(self, table: str, connection: str = None) -> str:
        """Get a SQLAlchemy tablename with prefix by name (without prefix) and connection str or connection.tablename dot notation"""
        pass

    @abstractmethod
    def engine(self, connection: str = None, metakey: str = None) -> sa.engine.Engine:
        """Get one SQLAlchemy Engine by connection str or metakey"""
        pass

    @abstractmethod
    async def database(self, connection: str = None, metakey: str = None) -> EncodeDatabase:
        """Get one Encode Database by connection str or metakey"""
        pass

    @abstractmethod
    async def disconnect(self, connection: str = None, metakey: str = None) -> None:
        """Disconnect from database by connection str or metakey"""
        pass

    @abstractmethod
    async def disconnect_all(self):
        """Disconnect from all connected databases"""
        pass

    @abstractmethod
    async def fetchall(self, query: Union[ClauseElement, str], values: Dict = None, connection: str = None, metakey: str = None) -> List[Mapping]:
        """Fetch List of records from a SQLAlchemy Core Query based on connection str or metakey"""
        pass

    @abstractmethod
    async def fetchone(self, query: Union[ClauseElement, str], values: Dict = None, connection: str = None, metakey: str = None) -> Optional[Mapping]:
        """Fetch one record from a SQLAlchemy Core Query based on connection str or metakey"""
        pass

    @abstractmethod
    async def execute(self, query: Union[ClauseElement, str], values: Union[List, Dict] = None, connection: str = None, metakey: str = None) -> Any:
        """Execute a SQLAlchemy Core Query based on connection str or metakey"""
        pass

    @abstractmethod
    def query(self, connection: str = None) -> DbQueryBuilder[DbQueryBuilder, None]:
        """Database query builder passthrough"""

    # @property
    # @abstractmethod
    # def events(self) -> Dict: pass

    # @property
    # @abstractmethod
    # def listeners(self) -> Dict[str, List]: pass

    # @property
    # @abstractmethod
    # def wildcards(self) -> List: pass

    # @abstractmethod
    # def register(self, events: Dict):
    #     pass

    # @abstractmethod
    # def listen(self, events: Union[str, List], listener: Any) -> None:
    #     pass

    # @abstractmethod
    # def dispatch(self, event: Any, payload = {}) -> None:
    #     pass

    # @abstractmethod
    # def get_listeners(self, event: str) -> List:
    #     pass



import uvicore
from uvicore import log
from uvicore.support.dumper import dd, dump
from uvicore.console import command, argument

# Commands
#list = typer.Typer()
#show = typer.Typer()


@command()
def list():
    """List all events"""
    log.header("Events defined from all packages")
    log.line()
    dump(uvicore.events.events)


@command()
@argument('event')
def get(event: str):
    """Show detailed info for one event"""
    log.header("Event details for " + event)
    log.line()
    event = uvicore.events.event(event)
    if event:
        dump(event)
    else:
        typer.echo(f"Event {event} not found")



import uvicore
from uvicore import log
from faker import Faker
from xx_vendor.xx_appname.models import xx_ModelName
from uvicore.support.dumper import dump, dd


@uvicore.seeder()
async def seed():
    log.item('Seeding xx_modelname')
    xx_modelname_items = []
    fake = Faker()
    for _ in range(2):
        title = fake.text(max_nb_chars=50)
        xx_modelname = xx_ModelName(
            slug=fake.slug(title),
            title=title,
            creator_id=1,
        )
        xx_modelname_items.append(xx_modelname)

    # Bulk insert
    await xx_ModelName.insert(xx_modelname_items)



import uvicore
from uvicore.typing import Dict, Any
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.support.module import load, location
from uvicore.support.dictionary import deep_merge
from uvicore.console.provider import Cli
from uvicore.http.provider import Http



@uvicore.provider()
class Http(ServiceProvider, Cli, Http):

    def register(self) -> None:
        """Register package into uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs and IoC bindings.  Configs are deep merged only after
        all packages are registered.  No real work should be performed here as it
        is very early in the bootstraping process and most internal processes are not
        instantiated yet.
        """

        # Register IoC bindings only if running in HTTP mode
        if self.app.is_http:

            # # Bind HTTP Server
            # self.bind('Http', 'uvicore.http.server._Server',
            #     aliases=['uvicore.http.server.Server', 'http', 'HTTP'],
            #     singleton=True,
            #     kwargs={
            #         'debug': uvicore.config('app.debug'),
            #         'title': uvicore.config('app.openapi.title'),
            #         'version': uvicore.app.version,
            #         'openapi_url': uvicore.config('app.openapi.url'),
            #         'docs_url': uvicore.config('app.openapi.docs_url'),
            #         'redoc_url': uvicore.config('app.openapi.redoc_url'),
            #     }
            # )
            # No because I added default to make
            #self.bind('WebRouter', 'uvicore.http.routing.web_router._WebRouter', aliases=['uvicore.http.routing.web_router.WebRouter', 'web_router'])
            #self.bind('ApiRouter', 'uvicore.http.routing.api_router._ApiRouter', aliases=['uvicore.http.routing.api_router.ApiRouter', 'api_router'])
            #self.bind('Routes', 'uvicore.http.routing.routes._Routes', aliases=['uvicore.http.routing.routes.Routes', 'routes'])
            #self.bind('StaticFiles', 'uvicore.http.static._StaticFiles', aliases=['Static', 'static'])

            # Default templating system is Jinja2.  Users can overwrite this
            # easily in their app configs 'bindings' dictionary.
            #self.bind('Templates', 'uvicore.http.templating.jinja._Jinja', singleton=True, aliases=['templates'])

            # Set app instance variables
            self.app._http = uvicore.ioc.make('uvicore.http.server._Server')

            # Register event listeners
            # After all providers are booted we have a complete list of view paths
            # and template options fully merged.  Now we can fire up the static
            # paths and template system.
            self.events.listen('uvicore.foundation.events.app.Booted', self.booted)

    def boot(self) -> None:
        """Bootstrap package into uvicore framework.
        Boot takes place after all packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all configs.
        This is where you load views, assets, routes, commands...
        """
        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define HTTP Middleware
        #self.middleware(uvicore.config('app.middleware'))

        # Define CLI commands
        self.commands({
            'http': {
                'help': 'Uvicore HTTP Commands',
                'commands': {
                    'serve': 'uvicore.http.commands.serve.cli',
                },
            }
        })

        # Uvicore default template options.  Because this HTTP package is high up
        # your package can easily override any of these default options!
        from uvicore.http.templating import context_functions
        self.template({
            'context_functions': {
                'url': context_functions.url,
                'asset': context_functions.asset,
            },
        })

    def booted(self, event: str, payload: Any) -> None:
        """Custom event handler for uvicore.foundation.events.app.Booted"""
        # This function only runs if in HTTP mode, not CLI mode.
        # So no need to check if running as HTTP in each function below!
        # But we'll double check here just in case.
        if self.app.is_console: return

        # Loop each package with an HTTP definition and add to our HTTP server
        asset_paths = []
        view_paths = []
        template_options = {}
        for package in self.app.packages.values():
            if not 'http' in package: continue

            # Add web and API routes for this one package
            self.add_web_routes(package)
            self.add_api_routes(package)

            # Append asset paths for later
            for path in package.http.asset_paths or []:
                path_location = location(path)
                if path_location not in asset_paths: asset_paths.append(path_location)

            # Append template paths
            for path in package.http.view_paths or []:
                view_paths.append(path)

            # Deep merge template options
            template_options = deep_merge(package.http.template_options or {}, template_options)

        # Mount all static asset paths
        self.mount_static_assets(package, asset_paths)

        # Initialize new template environment from the options above
        self.initialize_templates(view_paths, template_options)

    def add_web_routes(self, package) -> None:
        """Import and load this one packages web routes"""
        # Dont load items if registration is disabled
        if not package.registers.web_routes: return

        # Do not load if no web_routes defined
        if not 'web_routes' in package.http: return

        # Import and instantiate apps WebRoutes class
        from uvicore.http.routing.web_router import WebRouter
        WebRoutes = load(package.http.web_routes).object
        routes = WebRoutes(uvicore.app, package, WebRouter, package.http.web_route_prefix)
        routes.register()

    def add_api_routes(self, package) -> None:
        """Import and load this one packages api routes"""
        # Dont load items if registration is disabled
        if not package.registers.api_routes: return

        # Do not load if no web_routes defined
        if not 'api_routes' in package.http: return

        # Import and instantiate apps ApiRoutes class
        from uvicore.http.routing.api_router import ApiRouter
        ApiRoutes = load(package.http.api_routes).object
        routes = ApiRoutes(uvicore.app, package, ApiRouter, package.http.api_route_prefix)
        routes.register()

    def mount_static_assets(self, package, asset_paths) -> None:
        """Mount /static route using all packages static paths"""
        StaticFiles = uvicore.ioc.make('uvicore.http.static._StaticFiles')

        # Mount all directories to /assets in one go
        # Last directory defined WINS, which fits our last provider wins
        self.app.http.mount('/assets', StaticFiles(directories=asset_paths), name='assets')

    def initialize_templates(self, paths, options) -> None:
        """Initialize template system"""
        # Get the template singleton from the IoC
        templates = uvicore.ioc.make('uvicore.http.templating.jinja._Jinja')

        # Add all packages view paths
        for path in paths:
            templates.include_path(location(path))

        # Add all packages deep_merged template options
        if 'context_functions' in options:
            for name, method in options['context_functions'].items():
                templates.include_context_function(name, method)
        if 'context_filters' in options:
            for name, method in options['context_filters'].items():
                templates.include_context_filter(name, method)
        if 'filters' in options:
            for name, method in options['filters'].items():
                templates.include_filter(name, method)
        if 'tests' in options:
            for name, method in options['tests'].items():
                templates.include_test(name, method)

        # Initialize template system
        templates.init()




from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List, TypeVar

from uvicore.contracts import Application, Package

# Generic Router (APIRouter or WebRouter)
R = TypeVar('R')

class Routes(Generic[R], ABC):

    @property
    @abstractmethod
    def app(self) -> Application: pass

    @property
    @abstractmethod
    def package(self) -> Package: pass

    @property
    @abstractmethod
    def Router(self) -> R: pass

    @property
    @abstractmethod
    def prefix(self) -> str: pass

    @abstractmethod
    def include(self, module, *, prefix: str = '', tags: List[str] = None) -> None:
        """Include a new router object"""
        pass



from abc import ABC, abstractmethod
from typing import List, Any, Type, Callable

from starlette.routing import BaseRoute

# class Router(ABC):

#     @abstractmethod
#     def package(self, name: str = None, module: str = None, main: bool = False):
#         pass

class WebRouter(ABC):
    """Router for Web Controllers"""

    @property
    @abstractmethod
    def router(self) -> Any: pass

    @property
    @abstractmethod
    def routes(self) -> List[BaseRoute]: pass

    @property
    @abstractmethod
    def on_startup(self) -> None: pass

    @property
    @abstractmethod
    def on_shutdown(self) -> None: pass

    @abstractmethod
    def get(self,
        path: str,
        name: str = None,
    ) -> Callable: pass

    @abstractmethod
    def include_router(self, router: "WebRouter") -> None: pass


class ApiRouter(ABC):
    """Router for API Endpoints"""

    @property
    @abstractmethod
    def router(self) -> Any: pass

    @property
    @abstractmethod
    def routes(self) -> List[BaseRoute]: pass

    @property
    @abstractmethod
    def on_startup(self) -> None: pass

    @property
    @abstractmethod
    def on_shutdown(self) -> None: pass

    @abstractmethod
    def get(self,
        path: str,
        name: str = None,
        *,
        response_model: Type[Any] = None,
        include_in_schema: bool = True
    ) -> Callable: pass

    @abstractmethod
    def include_router(self, router: "APIRouter") -> None: pass



from __future__ import annotations

from typing import Optional

import uvicore
from uvicore.auth.database.tables import user_info as table
#from uvicore.orm.fields import BelongsTo, Field
#from uvicore.orm.model import Model, ModelMetaclass
from uvicore.support.dumper import dd, dump
from uvicore.orm import Model, ModelMetaclass, Field, BelongsTo

#@uvicore.ioc.bind('uvicore.auth.models.user_info.UserInfo')
#class UserInfoModel(Model['UserInfoModel'], metaclass=ModelMetaclass):

@uvicore.model()
class UserInfo(Model['UserInfo'], metaclass=ModelMetaclass):
#class UserModel(Model['UserModel']):
    """Auth User Model"""

    # Database connection and table information
    __tableclass__ = table.UserInfo

    id: Optional[int] = Field('id',
        primary=True,
        description='User Primary ID',
        sortable=True,
        searchable=True,
    )

    extra1: str = Field('extra1',
        description='Extra1',
    )

    user_id: int = Field('user_id',
        description="Contacts User ID",
        required=True,
    )

    # One-To-One Inverse - Contact has ONE User
    user: Optional[User] = Field(None,
        description="User Model",
        #belongs_to=('uvicore.auth.models.user.User', 'id', 'user_id'),
        #relation=BelongsTo('uvicore.auth.models.user.User', 'id', 'user_id')
        relation=BelongsTo('uvicore.auth.models.user.User')
    )

# IoC Class Instance
#UserInfo: UserInfoModel = uvicore.ioc.make('uvicore.auth.models.user_info.UserInfo', UserInfoModel)
#class UserInfo(UserInfoIoc, Model[UserInfoModel], UserInfoInterface): pass

# class UserInfo(
#     _UserInfo,
#     Model[UserInfoModel],
#     UserInfoInterface
# ): pass


#from uvicore.auth.models.user import UserModel as User  # isort:skip
from uvicore.auth.models.user import User  # isort:skip
#User = uvicore.ioc.make('uvicore.auth.models.user.User')
UserInfo.update_forward_refs()



from uvicore.typing import Any, AsyncGenerator, Dict, List, Mapping, Optional, Union

import sqlalchemy as sa
from databases import Database as EncodeDatabase
from sqlalchemy.sql import ClauseElement

import uvicore
from uvicore.contracts import Connection
from uvicore.contracts import Database as DatabaseInterface
from uvicore.database.query import _DbQueryBuilder
from uvicore.support.dumper import dd, dump
from sqlalchemy.engine.result import RowProxy

@uvicore.service('uvicore.database.db._Db',
    aliases=['Database', 'database', 'db'],
    singleton=True,
)
class _Db(DatabaseInterface):
    """Database private class.

    Do not import from this location.
    Use the uvicore.db singleton global instead."""

    @property
    def default(self) -> str:
        return self._default

    @property
    def connections(self) -> Dict[str, Connection]:
        return self._connections

    @property
    def engines(self) -> Dict[str, sa.engine.Engine]:
        return self._engines

    @property
    def databases(self) -> Dict[str, EncodeDatabase]:
        return self._databases

    @property
    def metadatas(self) -> Dict[str, sa.MetaData]:
        return self._metadatas

    def __init__(self) -> None:
        self._default = None
        self._connections = Dict()
        self._engines = Dict()
        self._databases = Dict()
        self._metadatas = Dict()

    def init(self, default: str, connections: Dict[str, Connection]) -> None:
        self._default = default
        self._connections = connections

        # For each unique metakey, create engines, encode databases and metadatas
        for connection in connections.values():
            # Check if we have already handled this unique metakey
            if connection.metakey in self.metadatas: continue

            # Build encode/databases specific connection URL
            # connection.url has a dialect in it, which we need for engines
            # but don't need for encode/databases library
            if connection.driver == 'sqlite':
                encode_url = (connection.driver
                    + ':///' + connection.database
                )
            else:
                encode_url = (connection.driver
                    + '://' + connection.username
                    + ':' + connection.password
                    + '@' + connection.host
                    + ':' + str(connection.port)
                    + '/' + connection.database
                )
            self._engines[connection.metakey] = sa.create_engine(connection.url)
            self._databases[connection.metakey] = EncodeDatabase(encode_url)
            self._metadatas[connection.metakey] = sa.MetaData()

        if uvicore.app.is_http:
            @uvicore.app.http.on_event("startup")
            async def startup():
                for database in self.databases.values():
                    await database.connect()

            @uvicore.app.http.on_event("shutdown")
            async def shutdown():
                for database in self.databases.values():
                    await database.disconnect()

    def packages(self, connection: str = None, metakey: str = None) -> Connection:
        if not metakey:
            if not connection: connection = self.default
            metakey = self.connection(connection).metakey
        packages = []
        for package in uvicore.app.packages.values():
            if not 'database' in package: continue
            for conn in package.database.connections.values():
                if conn.metakey == metakey:
                    packages.append(package)
        return packages

    def metakey(self, connection: str = None, metakey: str = None) -> str:
        try:
            if not metakey:
                if not connection:
                    connection = self.default
                metakey = self.connection(connection).metakey
            return metakey
        except Exception:
            raise Exception('Metakey not found, connection={} metakey={}'.format(connection, metakey))

    def connection(self, connection: str = None) -> Connection:
        if not connection: connection = self.default
        return self.connections.get(connection)

    def metadata(self, connection: str = None, metakey: str = None) -> sa.MetaData:
        metakey = self.metakey(connection, metakey)
        return self.metadatas.get(metakey)

    def table(self, table: str, connection: str = None) -> sa.Table:
        tablename = self.tablename(table, connection)
        metadata = self.metadata(connection)
        if metadata: return metadata.tables.get(tablename)

    def tablename(self, table: str, connection: str = None) -> str:
        if '.' in table:
            connection, table = tuple(table.split('.'))
        connection = self.connection(connection)
        if connection:
            return connection.prefix + table

    def engine(self, connection: str = None, metakey: str = None) -> sa.engine.Engine:
        metakey = self.metakey(connection, metakey)
        return self.engines.get(metakey)

    async def database(self, connection: str = None, metakey: str = None) -> EncodeDatabase:
        metakey = self.metakey(connection, metakey)
        database = self.databases.get(metakey)
        if not database.is_connected:
            await database.connect()
        #await self._connect(metakey=metakey)
        return self.databases.get(metakey)

    async def disconnect(self, connection: str = None, metakey: str = None) -> None:
        metakey = self.metakey(connection, metakey)
        database = self.databases.get(metakey)
        if database.is_connected:
            await database.disconnect()

    async def disconnect_all(self):
        for database in self.databases.keys():
            await self.disconnect(metakey=database)

    async def fetchall(self, query: Union[ClauseElement, str], values: Dict = None, connection: str = None, metakey: str = None) -> List[RowProxy]:
        database = await self.database(connection, metakey)
        return await database.fetch_all(query, values)

    async def fetchone(self, query: Union[ClauseElement, str], values: Dict = None, connection: str = None, metakey: str = None) -> Optional[RowProxy]:
        database = await self.database(connection, metakey)
        return await database.fetch_one(query, values)

    async def execute(self, query: Union[ClauseElement, str], values: Union[List, Dict] = None, connection: str = None, metakey: str = None) -> Any:
        database = await self.database(connection, metakey)
        if type(values) == dict:
            return await database.execute(query, values)
        elif type(values) == list:
            return await database.execute_many(query, values)
        else:
            return await database.execute(query)

    #  async def iterate(self, query: Union[ClauseElement, str], values: dict = None, connection: str = None) -> AsyncGenerator[Mapping, None]:
    #     async with self.connection() as connection:
    #         async for record in connection.iterate(query, values):
    #             yield record

    # async def _connect(self, connection: str = None, metakey: str = None) -> None:
    #     # Async connect to db if not connected
    #     # If running from web, we will already be connected from on_event("startup")
    #     # but if from CLI, we wont
    #     metakey = self.metakey(connection, metakey)
    #     database = self.databases.get(metakey)  # Dont call self.database() as its recursive
    #     if not database.is_connected:
    #         await database.connect()


    def query(self, connection: str = None) -> _DbQueryBuilder[_DbQueryBuilder, Any]:
        if not connection: connection = self.default
        return _DbQueryBuilder(connection)


    # def table(self, table: str):
    #     return QueryBuilder(self, self.default, table)

    # def conn(self, connection: str):
    #     return QueryBuilder(self, connection)


# IoC Class Instance
# Not to be imported by the public from here.
# Use the uvicore.db singleton global instead.
















#class QueryBuilder:
#     def __init__(self, db: DatabaseInterface, connection: str, table: str = None):
#         self.db = db
#         self.connection = connection or self.db.default
#         self.prefix = self.db.connection(self.connection).prefix
#         self._table = table
#         self._join = None
#         self._select = None
#         self._where = []

#     def table(self, table: str):
#         self._table = self.prefix + table
#         return self

#     def select(self, *args):
#         self._select = args
#         return self

#     def join(self, table: str, left: str, operator: str, right: str):
#         dd('hi')

#     def where(self, column: str, operator: str = None, value: Any = None):
#         if not value:
#             value = operator
#             operator = '='
#         self._where.append((column, operator, value))
#         return self

#     # Finalizers are find, get, update, delete, insert

#     async def find(self, pk: Any):
#         # FIDME, need dynamic pk
#         table, query = self.build()
#         query = query.where(table.c.id == pk)
#         return await self.db.fetchone(query)

#     async def get(self):
#         table, query = self.build()

#         # if self._select is none
#         # query = sa.select([table.c.id, table.c.email])
#         # #query = table.select()

#         #x = [getattr(table.c, x) for x in self._select]
#         #return x

#         return await self.db.fetchall(query)
#         return {
#             'db': self.db,
#             'connection': self.db.connection(self.connection),
#             'select': self._select,
#             'table': table,
#             'where': self._where,
#         }

#     def build(self, method: str = 'select'):
#         table = self.get_table()
#         query = None

#         # where
#         #   select
#         #   update (not in insert)
#         #   delete

#         if method == 'select':
#             if self._select:
#                 columns = []
#                 #for select in self._select:
#                 columns = [getattr(table.c, x) for x in self._select]
#                 query = sa.select(columns)
#             else:
#                 query = sa.select([table])

#         return table, query

#     def get_table(self):
#         #return self.db.table(self._table, connection=self.connection)
#         metadata = self.db.metadata(self.connection)
#         if metadata:
#             return metadata.tables.get(self._table)

#adf



import uvicore
from uvicore.typing import List, Any
from uvicore.http.routing.api_router import ApiRouter
from uvicore.typing import TypeVar, Generic, OrderedDict, Dict
from uvicore.support import str as string
from uvicore.support.dumper import dump, dd

E = TypeVar("E")

#from app1.models import Hashtag

class ModelRoute(Generic[E]):
    def __init__(self, Model: E, router: ApiRouter, path: str):
        self.Model: E = Model
        self.router: ApiRouter = router
        self.path: str = path

    def routes(self):
        tags = [string.ucbreakup(self.path)]

        # GET List
        self.router.add_api_route(
            path='/' + self.path,
            methods=['GET'],
            tags=tags,
            endpoint=self.list,
            response_model=List[self.Model],
            name='Get all {}'.format(self.path),
            #name=self.path,
            #   name='',
        )

        # GET
        self.router.add_api_route(
            path='/' + self.path + '/{id}',
            methods=['GET'],
            tags=tags,
            endpoint=self.get,
            response_model=self.Model,
            name='Get {} by ID'.format(self.path),
            #name=self.path,
            #name='',
        )

        # POST
        self.router.add_api_route(
            path='/' + self.path,
            methods=['POST'],
            tags=tags,
            endpoint=self.post,
            response_model=self.Model,
            name='Create {}'.format(self.path),
            #name=self.path,
            #name='',
        )

    async def list(self, include: str = ''):
        return await self.Model.query().include(*include.split(',')).get()

    async def get(self, id: Any, include: str = ''):
        return await self.Model.query().include(*include.split(',')).find(id)

    async def post(self, entity):
        dump(entity)
        return entity
        #return await self.Model.query().include(*include.split(',')).find(id)



@uvicore.service()
class ModelRouter:  # Need Interface

    def routes(self):

        # New Api Router
        router = ApiRouter()

        # Get all models in tablename sorted order
        models = self._get_models()

        # Loop each sorted model and dynamically add a ModelRouter
        for key, binding in models.items():
            # Get actual model from Ioc binding object
            Model = binding.object

            # Get URL path
            path = Model.tablename

            # Add model routes
            ModelRoute[Model](Model, router, path).routes()

        # Return router
        #for r in router.routes: dump(r.__dict__)
        #exit()
        #dd(route.routes)
        return router

    def _get_models(self) -> OrderedDict:
        """Get all models in tablename sorted order"""
        models = OrderedDict()
        unsorted_models = {}
        model_bindings = uvicore.ioc.binding(type='model', include_overrides=False)
        for key, binding in model_bindings.items():
            tablename = binding.object.tablename
            if tablename not in models:
                unsorted_models[tablename] = binding

        # Sort the OrderedDict
        #import operator
        #x = sorted(models.items(), key=lambda x: x[1])
        #x = sorted((value, key) for (key,value) in models.items())
        keys = sorted(unsorted_models.keys())
        for key in keys:
            models[key] = unsorted_models[key]
        return models



import uvicore
from uvicore.console import click, group


title_ideas = """
    A Fast Async Python Framework for CLI, Web and API

    High Performance Web, API and CLI Python Framework

    Python async Web, API and CLI fullstack Framework

    Fullstack Asynchronous Python Framework
    Python web framework
    Python Fullstack Asynchronous Web, API and CLI Framework

    Asynchronous Python Web, API and CLI Fullstack Framework

    The Async Python Framework for Craftsmen
    The Async Python Framework for Artisans
    The Artisanal Asynchronous Python Framework
"""

# Must include actual bind name when stacking decorators as it cannot deduce on its own.
# Bind decorator must come first in the stack
@uvicore.service('uvicore.console.console._cli', aliases=['Console', 'console', 'cli'])
@group(help=f"""
    \b
    Uvicore {uvicore.__version__}
    The Async Python Framework for Artisans.
    An Elegant Fullstack Python Web, API and CLI Framework

    Copyright (c) 2020 Matthew Reschke
    License http://mreschke.com/license/mit
""")
@click.version_option(version=uvicore.__version__, prog_name='Uvicore Framework', flag_value='--d')
async def _cli():
    await before_command()
    pass


async def before_command():
    # FIXME, right here you can perform BEFORE any command code, like before_command event dispatch
    # Could also do this in the command decorator?  No because all commands are loaded up front, so all fire
    # before a single command does
    #print('before command here')
    pass


@_cli.resultcallback()
async def after_command(result, **kwargs):
    # FIXME, right here you can perform AFTER any command code, like after_command event dispatch

    # Disconnect database
    # Will fail if uvicore.database package is not loaded, so check first
    if 'uvicore.database' in uvicore.app.providers:
        try:
            await uvicore.db.disconnect_all()
        except:
            pass

# IoC Class Instance
# No because not to be used by the public



import uvicore
from uvicore.console import command, argument, option
from uvicore.support.dumper import dump, dd


@command()
async def command():
    """A New CLI Command"""
    print('New command here')



import uvicore
from typing import Dict
from uvicore.contracts import Database as DatabaseInterface
from uvicore.support.dumper import dd, dump
from uvicore.contracts import Connection
from sqlalchemy import create_engine
from sqlalchemy import MetaData


class _Db(DatabaseInterface):

    # Class Variables
    default: str = None
    connections: Dict[str, Connection] = {}
    engines: Dict[str, str] = {}
    metadata: Dict[str, str] = {}

    @staticmethod
    def add_connection(connection: Connection) -> None:
        __class__.connections[connection.name] = connection
        __class__.engines[connection.name] = create_engine(connection.url)
        __class__.metadata[connection.name] = MetaData()

    # Instance Variables
    # @property
    # def default(self) -> str:
    #     return self._default

    # @property
    # def connections(self) -> Dict[str, Connection]:
    #     return self._connections

    # @property
    # def engines(self) -> Dict[str, str]:
    #     return self._engines

    # @property
    # def metadata(self) -> Dict[str, str]:
    #     return self._metadata

    # @default.setter
    # def default(self, value: str) -> None:
    #     self._default = value

    def __init__(self, connection: str = None):
        if not connection:
            connection = __class__.default
        #self._default = None
        #self._connections = {}
        #self._engines = {}
        #self._metadata = {}
        self._connection = connection
        self._table = None
        self._where = None

    def table(self, name: str):
        self._table = name
        return self

    def get(self):
        con = self._connect()
        query = self._query()
        results = con.execute(query)
        return results
        # rows = []
        # for row in results:
        #     rows.append(row)
        #     #rows.append(entity.User(**row))
        #     #rows.append(entity(**row))
        # return rows

    def _query(self):
        """Convert query builder into SQLAlechemy query"""
        #dd(__class__.metadata.get(self._connection).tables['users'])
        table = self._sa_table(self._table)
        query = table.select()
        return query

    def _sa_table(self, name: str):
        tables = __class__.metadata.get(self._connection).tables
        if name in tables:
            return tables.get(name)

    def _connect(self):
        return self.engine().connect()

    def engine(self):
        return __class__.engines.get(self._connection)


    # def _fetchall(self, entity, query) -> List[E]:
    #     if app.is_async:
    #         return self.encodedb.fetch_all(query=query)
    #     else:
    #         with self.connect() as con:
    #             results = con.execute(query)
    #             rows = []
    #             for row in results:
    #                 #rows.append(entity.User(**row))
    #                 rows.append(entity(**row))
    #             return rows


    # def _build_connection(self):
    #     if not 'connection' in self._query:
    #         self._query['connection'] = self.default




    # def add_connection(self, name: str, driver: str,
    #     dialect: str, host: str, port: int, database: str,
    #     username: str, password: str, prefix: str
    # ) -> None:
    #     url = (driver
    #         + '+' + dialect
    #         + '://' + username
    #         + ':' + password
    #         + '@' + host
    #         + ':' + str(port)
    #         + '/' + database)

    #     self._connections[name] = {
    #         'name': name,
    #         'driver': driver,
    #         'dialect': dialect,
    #         'host': host,
    #         'port': port,
    #         'url': url,
    #         'prefix': prefix
    #     }


# IoC Class Instance
# No, not a public API

# Public API for import * and doc gens
__all__ = ['_Db']



import uvicore
from uvicore.typing import Any, Dict
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.support.module import load
from uvicore.console.provider import Cli


@uvicore.provider()
class Database(ServiceProvider, Cli):

    def register(self) -> None:
        # Register IoC bindings
        # self.bind('Database', 'uvicore.database.db._Db',
        #     singleton=True,
        #     aliases=['database', 'db'],
        # )

        # Set uvicore.log global
        uvicore.db = uvicore.ioc.make('uvicore.database.db._Db')

        # Register event listeners
        # After all providers are registered (meaning configs merged)
        # Init the database system with all connection information
        #self.events.listen('uvicore.foundation.events.app.Registered', self.registered)

        # After all providers are booted, fire up database with all package
        # connections, modules, tables, seeders...
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)

    def boot(self) -> None:

        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define db commands
        self.commands(
            group='db',
            help='Database Commands',
            commands={
                'create': 'uvicore.database.commands.db.create',
                'drop': 'uvicore.database.commands.db.drop',
                'recreate': 'uvicore.database.commands.db.recreate',
                'seed': 'uvicore.database.commands.db.seed',
                'reseed': 'uvicore.database.commands.db.reseed',
                'connections': 'uvicore.database.commands.db.connections',
            }
        )

        # Extend schematic generator commands
        self.commands(
            group='gen',
            commands={
                'table': 'uvicore.database.commands.generators.table',
                'seeder': 'uvicore.database.commands.generators.seeder',
            }
        )

    def registered_OBSOLETE(self, event: str, payload: Any):
        """Custom event handler for uvicore.foundation.events.app.Registered"""
        default = uvicore.app.package(main=True).connection_default
        connections = []
        for package in uvicore.app.packages.values():
            for connection in package.connections:
                connections.append(connection)

        # Init the database with all connections
        uvicore.db.init(default, connections)

        # # Set default connection based on MAIN package
        # uvicore.db.default = uvicore.app.package(main=True).connection_default

        # # Add each packages connection
        # for package in uvicore.app.packages.values():
        #     for connection in package.connections:
        #         uvicore.db.add_connection(connection)

    def booted(self, event: str, payload: Any) -> None:
        """Custom event handler for uvicore.foundation.events.app.Booted"""

        # Gather all connections, models, tables and seeders
        connections = Dict()
        models = []; tables = []; seeders = []
        last_default = None; app_default = None
        for package in self.app.packages.values():
            if not 'database' in package: continue

            # Get last defined default connection
            if package.database.connection_default: last_default = package.database.connection_default

            # Get running app default connection
            if package.main and package.database.connection_default: app_default = package.database.connection_default

            # Append connections
            #connections.extend(package.database.connections or [])
            connections.merge(package.database.connections)
            #connections.merge(package.database)

            # Append models
            models.extend(package.database.models or [])

            # Append tables
            tables.extend(package.database.tables or [])

            # Append seeders
            seeders.extend(package.database.seeders or [])

        # Initialize Database with all connections at once
        uvicore.db.init(app_default or last_default, connections)

        # Dynamically Import models, tables and seeders
        for model in models: load(model)
        for table in tables: load(table)
        for seeder in seeders: load(seeder)
        #dump(connections, models, tables, seeders, last_default, app_default)



import uvicore
from typing import Any, Dict, List

def table(name: str = None):
    def decorator(cls):
        # Bind this table into the Ioc
        bind_name = name or cls.__module__ + '.' + cls.__name__
        new_cls = uvicore.ioc.bind_from_decorator(cls, name=bind_name, object_type='table', singleton=True)

        # Other things with a table

        return new_cls
    return decorator



import os
import sys

import uvicore
from uvicore.typing import Any, List, NamedTuple, Tuple, Dict, OrderedDict
from uvicore.package import Package
from uvicore.contracts import Application as ApplicationInterface
from uvicore.contracts import Config as ConfigInterface
from uvicore.contracts import Package as PackageInterface
from uvicore.contracts import Server as ServerInterface
from uvicore.contracts import Template as TemplateInterface
from uvicore.database import Connection
from uvicore.support.collection import dotget
from uvicore.support.dumper import dd, dump
from uvicore.support.hash import md5
from uvicore.support.module import load, location


@uvicore.service('uvicore.foundation.application._Application',
    aliases=['Application', 'application', 'App', 'app'],
    singleton=True,
)
class _Application(ApplicationInterface):
    """Application private class.

    Do not import from this location.
    Use the uvicore.app singleton global instead."""

    # Instance Variables
    @property
    def version(self) -> str:
        return self._version

    @property
    def debug(self) -> bool:
        return self._debug

    @property
    def perfs(self) -> List:
        return self._perfs

    @property
    def http(self) -> ServerInterface:
        return self._http

    # No, don't want duplicate entrypoints
    # @property
    # def config(self) -> ConfigInterface:
    #     return self._config

    @property
    def providers(self) -> OrderedDict[str, Dict]:
        return self._providers

    @property
    def registered(self) -> bool:
        return self._registered

    @property
    def booted(self) -> bool:
        return self._booted

    @property
    def is_console(self) -> bool:
        return self._is_console

    @property
    def is_http(self) -> bool:
        return self._is_http

    @property
    def packages(self) -> OrderedDict[str, PackageInterface]:
        return self._packages

    @property
    def path(self) -> str:
        return self._path

    @property
    def name(self) -> str:
        return self._name

    @property
    def main(self) -> str:
        return self._main

    def __init__(self):
        # Instance variables
        self._version = uvicore.__version__
        self._debug = False
        self._perfs = []
        self._http = None
        #self._config = None  # None until config provider registered
        self._providers = OrderedDict()
        self._registered = False
        self._booted = False
        self._is_console = False
        self._is_http = False
        self._packages = OrderedDict()
        self._path = None
        self._name = None
        self._main = None

    def bootstrap(self, app_config: Dict, path: str, is_console: bool) -> None:
        """Bootstrap the uvicore application"""

        # Silently do not bootstrap multiple times
        if self.booted: return

        # App name and path
        self._path = path
        self._name = app_config.name
        self._main = app_config.main

        # Merge running config/app.py paths dictionary with defaults and full path
        self._build_paths(app_config)

        # Detect if running in console (to register commands)
        # Ensure console is False even when running ./uvicore http serve
        self._is_console = is_console
        if "'http', 'serve'" in str(sys.argv): self._is_console = False
        self._is_http = not self.is_console

        # Detect debug flag from main app config
        self._debug = app_config.debug

        # Build recursive providers graph
        self._build_provider_graph(app_config)

        # Failsafe if no http package, force console
        # This solves a ./uvicore http serve error if you don't have the http package
        if 'uvicore.http' not in self.providers:
            self._is_console = True
            self._is_http = False

        # Register and merge all providers
        self._register_providers(app_config)

        #dump(self.packages)
        #dd('REGISTERED')

        # Boot all providers
        #self._boot_providers()
        self._boot_providers(app_config)

        #dd(self.packages)

        # Return application
        return self

    def _build_provider_graph(self, app_config: Dict) -> None:
        """Build recursive dependency graph of all packages"""

        def recurse(package: str, options: Dict):
            # Get the config/package.py config
            package_config = self._get_package_config(package, options)

            # Get the packages service provider dependencies
            services = package_config.get('dependencies') or {}

            # Recurse into each service dependency and find those dependencies from their config/package.py configs
            for service, details in services.items():
                recurse(service, details)

            # Add to providers OrderedDict. Notice this will OVERWRITE if new provider defined last.
            # This gives the perfect LAST provider WINS!  Also because this is an OrderedDict the
            # last provider will overrite the value but the defined ORDER will remain the
            # same.  Perfect!
            self._providers[package] = options

        # Loop each main apps packages and recurse into each
        packages = app_config.get('packages') or {}
        for package, options in packages.items():
            recurse(package, options)

    def _register_providers(self, app_config: Dict) -> None:
        """Register all providers by calling each ServiceProviders register() method"""

        for package_name, service in self.providers.items():
            # Example:
            # package_name = uvicore.configuration
            # service = {'provider': 'uvicore.configuration.services.Configuration'}

            # Start a new package definition
            #x = OrderedDict()
            #x.dotset(package_name, package.Definition({
            self._packages[package_name] = Package({
                'name': package_name,
                'main': True if package_name == self.main else False,
                'path': location(package_name),
            })
            #self._packages.merge(x)
            #self._packages[package_name] = package.Definition({
            #dd(self.packages)

            # Instantiate the provider and call the register() method
            provider = load(service['provider']).object(
                app=self,
                name=package_name,
                package=None,  # Not available in register()
                app_config=app_config,
                package_config=self._get_package_config(package_name, service),
            )
            provider.register()

        # Complete registration
        self._registered = True
        uvicore.events.dispatch('uvicore.foundation.events.app.Registered')

    def _boot_providers(self, app_config: Dict) -> None:
        for package_name, service in self.providers.items():
            # Example:
            # package_name = uvicore.configuration
            # service = {'provider': 'uvicore.configuration.services.Configuration'}

            # Import the provider and call boot()
            #dd(self.package(main=True))
            #dd(self.package(package_name))
            provider = load(service['provider']).object(
                app=self,
                name=package_name,
                package=self.package(package_name),
                app_config=app_config,
                package_config=self._get_package_config(package_name, service),
            )
            provider.boot()

        # Complete booting
        self._booted = True
        uvicore.events.dispatch('uvicore.foundation.events.app.Booted')

    def _build_paths(self, app_config: Dict):
        base = self.main.replace('.', '/')
        defaults = {
            'base': base,
            'commands': 'commands',
            'config': 'config',
            'database': 'database',
            'migrations': 'database/migrations',
            'seeders': 'database/seeders',
            'tables': 'database/tables',
            'events': 'events',
            'http': 'http',
            'api': 'http/api',
            'assets': 'http/assets',
            'controllers': 'http/controllers',
            'routes': 'http/routes',
            'static': 'http/static',
            'views': 'http/views',
            'jobs': 'jobs',
            'listeners': 'listeners',
            'models': 'models',
            'services': 'services',
            'support': 'support',
        }
        #dd(app_config)
        app_config['paths'].defaults(defaults)
        # = {**defaults, **app_config['paths']}
        for key, value in app_config['paths'].items():
            app_config['paths'][key] = os.path.realpath(self.path + '/' + self.main.replace('.', '/') + '/' + value)

    def _get_package_config(self, package: str, options: Dict) -> Dict:
        config_module = package + '.config.package.config'  # Default if not defined
        if 'config' in options: config_module = options['config']
        config = Dict()
        try:
            config = Dict(load(config_module).object)
        except:
            # Often we won't have any config for a package, if so return empty Dict
            pass
        return config

    def package(self, package: str = None, *, main: bool = False) -> PackageInterface:
        if package:
            return self.packages.get(package)
            #return self.packages.dotget(package)
        elif main:
            return next(package for package in self.packages.values() if package.main == True)
            #return self.packages.dotget(self.main)

    def perf(self, item) -> None:
        if self.debug:
            self.perfs.append(item)
            print(item)

    # NO, don't want duplicates of everything everywhere, just import dumper
    # def dump(self, *args) -> None:
    #     dump(*args)

    # def dd(self, *args) -> None:
    #     dd(*args)


# IoC Class Instance
# **Not meant to be imported from here**.  Use the uvicore.app singleton global instead.
# Only here because uvicore bootstrap needs to import it without a service provider.
# By using the default bind and make feature of the IoC we can swap the implimentation
# at a high bootstrap level using our app configs 'bindings' dictionary.
# The only two classes that do this are Application and the event Dispatcher.
#Application: _Application = uvicore.ioc.make('Application', _Application, singleton=True, aliases=['App', 'app', 'application'])



from typing import Any, Callable, List, Type

from starlette.routing import BaseRoute
from starlette.routing import Router as _StarletteRouter

import uvicore
from uvicore.contracts import WebRouter as RouterInterface
from uvicore.support.dumper import dd, dump


# Example of just using the entier Starlette Router without abstraction
@uvicore.service()
class WebRouter(_StarletteRouter):

    # Add a GET route decorator to starlette
    def get(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self.route(
            path=path,
            name=name,
            methods=['GET'],
            include_in_schema=False,
        )

    # Add a POST route decorator to starlette
    def post(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['POST'],
            include_in_schema=False,
        )

    # Add a PUT route decorator to starlette
    def put(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['PUT'],
            include_in_schema=False,
        )

    # Add a PATCH route decorator to starlette
    def patch(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['PATCH'],
            include_in_schema=False,
        )

    # Add a DELETE route decorator to starlette
    def delete(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['DELETE'],
            include_in_schema=False,
        )

    # Add a OPTIONS route decorator to starlette
    def options(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['OPTIONS'],
            include_in_schema=False,
        )

    # Add a HEAD route decorator to starlette
    def head(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['HEAD'],
            include_in_schema=False,
        )

    # Add a TRACE route decorator to starlette
    def trace(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['TRACE'],
            include_in_schema=False,
        )



# Example of making my own router abstraction
class WebRouterXXX(RouterInterface):

    @property
    def router(self) -> _StarletteRouter:
        return self._router

    @property
    def routes(self) -> List[BaseRoute]:
        return self._router.routes

    @property
    def on_startup(self) -> None:
        return self._router.on_startup

    @property
    def on_shutdown(self) -> None:
        return self._router.on_shutdown

    def __init__(self):
        # Fireup Starlette Router
        self._router = _StarletteRouter()

    def get(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['GET']
        )

    def post(self,
        path: str,
        name: str = None,
    ) -> Callable:
        return self._router.route(
            path=path,
            name=name,
            methods=['POST']
        )

    # Not needed
    # def include_router(self, router: "WebRouter") -> None:
    #     # Starlette does not have a _router.include_router  like FastAPI
    #     # This manually adds each route in the router to starlette router
    #     for route in router.routes:
    #         self._router.routes.append(route)


# IoC Class Instance
#WebRouter: RouterInterface = uvicore.ioc.make('WebRouter', _WebRouter)

# Public API for import * and doc gens
#__all__ = ['WebRouter', '_WebRouter']



from __future__ import annotations

from typing import Optional

import uvicore
from uvicore.auth.database.tables import users as table
#from uvicore.orm.fields import Field, HasOne
#from uvicore.orm.model import Model, ModelMetaclass
from uvicore.support.dumper import dd, dump
from uvicore.orm import Model, ModelMetaclass, Field, HasOne

#Model: _Model = uvicore.ioc.make('Model', _Model)
#import sys
#sys.modules['uvicore.orm.model']._Model = 'asdf'
#dump(sys.modules['uvicore.orm.model'].__dict__)

#from uvicore.orm.model import _Model
#dump(_Model)
#asdf('asdf')

# from abc import ABC, abstractmethod
# class UserInterface(ABC):
#     # @abstractmethod
#     # def idx(self) -> Optional[int]:
#     #     pass

#     id2: Optional[int]


#from uvicore.contracts import Model as ModelInterface
# # Model interfaces, though redundant, are used for proper type hinting code intellisense
# class User(ModelInterface):
#     id: Optional[int]
#     email: str
#     info: Optional[List[UserInfo]]


# UserModel for typehints only.  Import User for actual usage.
#class UserModel(Model, metaclass=ModelMetaclass):


#@uvicore.ioc.bind('uvicore.auth.models.user.User')

@uvicore.model()
class User(Model['User'], metaclass=ModelMetaclass):
#class UserModel(Model['UserModel']):
    """Auth User Model"""

    # Database connection and table information
    __tableclass__ = table.Users

    id: Optional[int] = Field('id',
        primary=True,
        description='User Primary ID',
        sortable=True,
        searchable=True,
    )

    email: str = Field('email',
        description='User Email and Username',
        required=True,
    )

    # One-To-One - User has ONE Contact
    info: Optional[UserInfo] = Field(None,
        description='User Info Model',
        relation=HasOne('uvicore.auth.models.user_info.UserInfo', foreign_key='user_id'),
    )

    # class Config:
    #     extra = 'ignore'
    #     arbitrary_types_allowed = True


# IoC Class Instance
#User = UserModel
#User: UserModel = uvicore.ioc.make('uvicore.auth.models.user.User', UserModel)
#User: UserModel = uvicore.ioc.make('uvicore.auth.models.user.User')
#class User(UserIoc, Model[UserModel], UserInterface): pass

# class User(
#     _User,
#     Model[_User],
#     UserInterface
# ): pass



from uvicore.auth.models.user_info import UserInfo  # isort:skip
#UserInfo = uvicore.ioc.make('uvicore.auth.models.user_info.UserInfo')
User.update_forward_refs()



import uvicore
from typing import Dict, List
from uvicore.contracts import Connection
from uvicore.contracts import Database as DatabaseInterface
from uvicore.support.dumper import dd, dump
import sqlalchemy as sa
# from sqlalchemy import MetaData as SaMetaData
# from sqlalchemy.engine import Engine as SaEngine
# from sqlalchemy.engine import Connection as SaConnection


class _Db(DatabaseInterface):

    @property
    def default(self) -> str:
        return self._default

    @property
    def connections(self) -> Dict[str, Connection]:
        return self._connections

    @property
    def engines(self) -> Dict[str, str]:
        return self._engines

    @property
    def metadatas(self) -> Dict[str, sa.MetaData]:
        return self._metadatas

    @default.setter
    def default(self, value: str) -> None:
        self._default = value

    def __init__(self) -> None:
        self._default = None
        self._connections = {}
        self._engines = {}
        self._metadatas = {}

    def init(self, default: str, connections: List[Connection]) -> None:
        self._default = default
        for connection in connections:
            self._connections[connection.name] = connection
            self._engines[connection.metakey] = sa.create_engine(connection.url)
            self._metadatas[connection.metakey] = sa.MetaData()

    def packages(self, connection: str = None, metakey: str = None) -> Connection:
        """Get all packages with the metakey derived from the connection name
        or passed in metakey.
        """
        if not metakey:
            if not connection: connection = self.default
            metakey = self.connection(connection).metakey
        packages = []
        for package in uvicore.app.packages.values():
            for conn in package.connections:
                if conn.metakey == metakey:
                    packages.append(package)
        return packages

    def connection(self, connection: str = None) -> Connection:
        """Get one connection by connection name"""
        if not connection: connection = self.default
        return self.connections.get(connection)

    def metadata(self, connection: str = None, metakey: str = None) -> sa.MetaData:
        """Get one metadata by connection name or metakey"""
        if metakey:
            return self.metadatas.get(metakey)
        else:
            if not connection: connection = self.default
            metakey = self.connection(connection).metakey
            return self.metadatas.get(metakey)

    def engine(self, connection: str = None, metakey: str = None) -> sa.engine.Engine:
        """get one engine by connection name or metakey"""
        if metakey:
            return self.engines.get(metakey)
        else:
            if not connection: connection = self.default
            metakey = self.connection(connection).metakey
            return self.engines.get(metakey)

    def connect(self, connection: str = None, metakey: str = None) -> sa.engine.Connection:
        """Connect to one engine by connection name or metakey"""
        if metakey:
            return self.engine(metakey=metakey).connect()
        else:
            if not connection: connection = self.default
            return self.engine(connection).connect()

    def execute(self, entity, *args, **kwargs):
        conn = self.connect(entity.__connection__)
        return conn.execute(*args, **kwargs)

    def fetchone(self, entity, query):
        return self.execute(entity, query).fetchone()

    def fetchall(self, entity, query):
        return self.execute(entity, query).fetchall()






from .application import Application
from .builder import DbQueryBuilder, OrmQueryBuilder, QueryBuilder
from .config import Config
from .connection import Connection
from .database import Database
from .dispatcher import Dispatcher
from .field import Field
from .ioc import Binding, Ioc
from .logger import Logger
from .mapper import Mapper
from .model import Model
from .package import Package
from .provider import Provider
from .relation import Relation
from .router import ApiRouter, WebRouter
from .routes import Routes
from .server import Server
from .template import Template

# The contracts package uses a __init__.py because users will import these
# methods from their own apps and we want a nicer import which looks
# like - from uvicore.contracts import Model



from typing import Callable, List

from fastapi import FastAPI as _FastAPI
from starlette.types import ASGIApp

import uvicore
from uvicore.contracts import Server as ServerInterface
from uvicore.support.dumper import dd, dump

# This is our actual FastAPI Server
# Can only have one server, can't be both FastAPI and/or Starlette
# The FastAPI server inherits Starlette server but extends it with
# extra options for openapi, so we must use that as our base http app

# In classic FastAIP example, this is same as
# app = FastAPI(), but I call it app.http

# Bound from service provider instead of decorator


@uvicore.service('uvicore.http.server._Server',
    aliases=['http', 'HTTP'],
    singleton=True,
    kwargs={
        'debug': uvicore.config('app.debug'),
        'title': uvicore.config('app.openapi.title'),
        'version': uvicore.app.version,
        'openapi_url': uvicore.config('app.openapi.url'),
        'docs_url': uvicore.config('app.openapi.docs_url'),
        'redoc_url': uvicore.config('app.openapi.redoc_url'),
    },
)
class _Server(ServerInterface):
    """HTTP Server private class.

    Do not import from this location.
    Use the uvicore.app.http singleton global instead."""

    @property
    def server(self) -> _FastAPI:
        return self._server

    def __init__(self, debug: bool, title: str, version: str, openapi_url: str, docs_url: str, redoc_url: str):

        # Fireup FastAPI HTTP Core
        self._server = _FastAPI(
            debug=debug,
            title=title,
            version=version,
            openapi_url=openapi_url,
            docs_url=docs_url,
            redoc_url=redoc_url,
        )

    def include_router(self, router, *, prefix: str = '', tags: List[str] = None) -> None:
        # _server is a FastAPI application.  If you look at FastAPIs include_router
        # it automatically detects if the the actual Router is FastAPI or Starlette
        # and handles the including of the routers routes differently!
        self._server.include_router(
            router=router,
            prefix=prefix,
            tags=tags,
        )

    def mount(self, path: str, app: ASGIApp, name: str = None) -> None:
        self._server.mount(path=path, app=app, name=name)

    def on_event(self, event_type: str) -> Callable:
        return self._server.on_event(event_type)


# IoC Class Instance
# No because not to be used by the public
#Server: ServerInterface = uvicore.ioc.make('Http')

# Public API for import * and doc gens
#__all__ = ['_Server']



import uvicore
from uvicore.contracts import Application as ApplicationInterface

@uvicore.service()
class Handler:

    @property
    def app(self) -> ApplicationInterface:
        return self._app

    def __init__(self, app: ApplicationInterface):
        self._app = app



import uvicore
from typing import Any, Dict, List

def seeder(name: str = None):
    def decorator(cls):
        # Bind this seeder into the Ioc
        bind_name = name or cls.__module__ + '.' + cls.__name__
        new_cls = uvicore.ioc.bind_from_decorator(cls, name=bind_name, object_type='seeder', singleton=False)

        # Other things with a seeder

        return new_cls
    return decorator



import uvicore
from typing import Dict
from uvicore.package import ServiceProvider
from uvicore.support.dumper import dump, dd
from uvicore.console.provider import Cli


@uvicore.provider()
class Orm(ServiceProvider, Cli):

    def register(self) -> None:
        pass
        # Register IoC bindings
        # Automatic - self.bind('Model', 'uvicore.orm.model._Model', aliases=['model'])
        # Automatic - self.bind('ModelMetaclass', 'uvicore.orm.metaclass._ModelMetaclass')

        #self.bind('Model', 'uvicore.orm.model._Model', aliases=['model'])
        #self.bind('ModelMetaclass', 'uvicore.orm.metaclass._ModelMetaclass')

    def boot(self) -> None:
        # Define service provider registration control
        self.registers(self.package.config.registers)

        # Define commands
        self.commands({
            # Extend schematic generator commands
            'gen': {
                'commands': {
                    'model': 'uvicore.orm.commands.generators.model',
                },
            }
        })



from typing import Generic, List, TypeVar, Union

import uvicore
from uvicore.contracts import Application as ApplicationInterface
from uvicore.contracts import Package as PackageInterface
from uvicore.contracts import Routes as RoutesInterface
from uvicore.support.module import load

# Generic Router (APIRouter or WebRouter)
R = TypeVar('R')


@uvicore.service()
class Routes(RoutesInterface, Generic[R]):

    @property
    def app(self) -> ApplicationInterface:
        return self._app

    @property
    def package(self) -> PackageInterface:
        return self._package

    @property
    def Router(self) -> R:
        return self._Router

    @property
    def prefix(self) -> str:
        return self._prefix

    def __init__(self,
        app: ApplicationInterface,
        package: PackageInterface,
        Router: R,
        prefix: str
    ):
        self._app = app
        self._package = package
        self._Router = Router
        self._prefix = prefix

    def include(self, module, *, prefix: str = '', tags: List[str] = None) -> None:
        #self.http.controller(controller.route, prefix=self.prefix)

        if type(module) == str:
            # Using a string to point to an endpoint class controller
            controller = load(self.endpoints + '.' + module + '.route')
            uvicore.app.http.include_router(
                controller.object,
                prefix=self.prefix + str(prefix),
                tags=tags,
            )
        else:
            # Passing in an actual router class
            uvicore.app.http.include_router(
                module,
                prefix=self.prefix + str(prefix),
                tags=tags,
            )

    # def Router(self) -> R:
    #     return self._Router()


# IoC Class Instance
#Routes: RoutesInterface = uvicore.ioc.make('Routes', _Routes)

# Public API for import * and doc gens
#__all__ = ['Routes', '_Routes']



import sqlalchemy as sa

import uvicore
from uvicore.database import Table
from uvicore.support.dumper import dd, dump

# Get related tablenames with proper prefixes
users = uvicore.db.tablename('auth.users')


@uvicore.table()
class UserInfo(Table):

    # Actual database table name
    # Plural table names and singluar model names are encouraged
    # Do not add a package prefix, leave that to the connection config
    name = 'user_info'

    # Connection for this database from your config file
    connection = 'auth'

    # SQLAlchemy Table definition as a list (exclude name and metadata)
    # This will be converted into an actual SQLAlchemy Table() instance
    # See https://docs.sqlalchemy.org/en/13/core/schema.html
    schema = [
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("extra1", sa.String(length=50)),
        sa.Column('user_id', sa.Integer, sa.ForeignKey(f"{users}.id"), nullable=False),
    ]

    # Optional SQLAlchemy Table() instance kwargs
    schema_kwargs = {}


# IoC Class Instance
#UserInfo: _UserInfo = uvicore.ioc.make('uvicore.auth.database.tables.user_info.UserInfo', _UserInfo, singleton=True)



import uvicore
from fastapi import APIRouter as _FastAPIRouter
from uvicore.typing import Any, Type, List, Callable, Optional
from starlette.routing import BaseRoute
from uvicore.contracts import ApiRouter as RouterInterface
from starlette.responses import Response


# Example of just using the entier FastAPI Router without abstraction
@uvicore.service()
class ApiRouter(_FastAPIRouter):
    pass




# Example of making my own router abstraction
class ApiRouterX(RouterInterface):

    @property
    def router(self) -> _FastAPIRouter:
        return self._router

    # These properties have to exist because the FastAPI code looks for route.routes
    # or routes.response_class for example so this router abstraction must pass them through
    # Perhaps my own abstraction isn't a good idea?  Maybe just pass through fastAPI's router?
    @property
    def routes(self) -> List[BaseRoute]:
        return self._router.routes

    @property
    def response_class(self) -> Type[Response]:
        return self._router.response_class

    @property
    def default_response_class(self) -> Type[Response]:
        return self._router.default_response_class

    @property
    def on_startup(self) -> None:
        return self._router.on_startup

    @property
    def on_shutdown(self) -> None:
        return self._router.on_shutdown

    def __init__(self):
        # Fireup FastAPI Router
        self._router = _FastAPIRouter()

    def get(self,
        path: str,
        name: str = None,
        *,
        response_model: Type[Any] = None,
        include_in_schema: bool = True
    ) -> Callable:
        return self._router.get(
            path=path,
            name=name,
            response_model=response_model,
            include_in_schema=include_in_schema,
        )

    #def include_router(self, router: "APIRouter", *, prefix: str = '', tags: List[str] = None) -> None:
    # def include_router(self, router: "APIRouter") -> None:
    #     # Only used for the class based controller
    #     # return self._router.include_router(router,
    #     #     prefix=prefix,
    #     #     tags=tags
    #     # )
    #     return self.router.include_router(router)

    def add_api_route(self, path: str, endpoint: Callable, *,
        response_model: Optional[Type[Any]] = None,
        name: Optional[str] = None,
        # self,
        # path: str,
        # endpoint: Callable,
        # *,
        # response_model: Optional[Type[Any]] = None,
        # status_code: int = 200,
        # tags: Optional[List[str]] = None,
        # dependencies: Optional[Sequence[params.Depends]] = None,
        # summary: Optional[str] = None,
        # description: Optional[str] = None,
        # response_description: str = "Successful Response",
        # responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        # deprecated: Optional[bool] = None,
        # methods: Optional[Union[Set[str], List[str]]] = None,
        # operation_id: Optional[str] = None,
        # response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        # response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        # response_model_by_alias: bool = True,
        # response_model_exclude_unset: bool = False,
        # response_model_exclude_defaults: bool = False,
        # response_model_exclude_none: bool = False,
        # include_in_schema: bool = True,
        # response_class: Optional[Type[Response]] = None,
        # name: Optional[str] = None,
        # route_class_override: Optional[Type[APIRoute]] = None,
        # callbacks: Optional[List[APIRoute]] = None,
    ) -> None:
        return self.router.add_api_route(
            path=path,
            endpoint=endpoint,
            response_model=response_model,
            name=name,
            # status_code=status_code,
            # tags=tags,
            # dependencies=dependencies,
            # summary=summary,
            # description=description,
            # response_description=response_description,
            # responses=responses,
            # deprecated=deprecated,
            # methods=methods,
            # operation_id=operation_id,
            # response_model_include=response_model_include,
            # response_model_exclude=response_model_exclude,
            # response_model_by_alias=response_model_by_alias,
            # response
        )

    # @property
    # def router(self):
    #     return self._router



# IoC Class Instance
#ApiRouter: RouterInterface = uvicore.ioc.make('ApiRouter', _ApiRouter)

# Public API for import * and doc gens
#__all__ = ['ApiRouter', '_ApiRouter']



import uvicore


@uvicore.seeder()
async def seed():
    # Import seeders
    from . import users

    # Run seeders. Order is critical for ForeignKey dependencies
    #await groups.seed()
    await users.seed()



from __future__ import annotations

from typing import Optional

import uvicore
from uvicore.auth.database.tables import groups as table
#from uvicore.orm.fields import Field
#from uvicore.orm.metaclass import ModelMetaclass
#from uvicore.orm.model import Model
from uvicore.support.dumper import dd, dump

from uvicore.orm import Model, ModelMetaclass, Field

@uvicore.model()
class Group(Model['Group'], metaclass=ModelMetaclass):
    """Auth Group Model"""

    # Database connection and table information
    __tableclass__ = table.Groups

    id: Optional[int] = Field('id',
        primary=True,
        description='Group primary ID',
        sortable=True,
        searchable=True,
    )

    name: str = Field('name',
        description='Group Name',
        required=True,
    )

# IoC Class Instance
#Group: GroupModel = uvicore.ioc.make('uvicore.auth.models.group.Group', GroupModel)

# class Group(
#     _Group,
#     Model[_Group],
#     GroupInterface
# ): pass



import uvicore
from uvicore import log
from faker import Faker
from uvicore.auth.models.user import User
from uvicore.support.dumper import dump, dd

@uvicore.seeder()
async def seed():
    log.item('Seeding table users')
    users = []
    fake = Faker()
    for _ in range(2):
        user = User(
            email=fake.email()
        )
        await user.save()



from uvicore.support.dumper import dump, dd
from uvicore.support.module import location
from uvicore.database import Connection
from uvicore.typing import Dict, List


class Db:
    """Database Service Provider Mixin"""

    #def _add_db_definition(self, key, value):
        # if type(value) == list:
        #     if not self.package.database[key]: self.package.database[key] = []
        #     self.package.database[key].extend(value)
        # elif type(value) == dict:
        #     self.package.database[key].merge(value)
        # else:
        #     self.package.database[key] = value

        #self.package.database[key] = value

        # if 'database' not in self.package:
        #     self.package.database = Dict()
        # if type(value) == list:
        #     if key not in self.package.database: self.package.database = []
        #     self.package['database'][key].extend(value)
        # else:
        #     self.package['database'][key] = value

    def connections(self, config: Dict, default: str):
        #connections = []
        for name, connection in config.items():

            # Build URL and metakey
            # Metakey cannot be the connection name.  If 2 connections share the exact
            # same database (host, port, dbname) then they need to also share the same
            # metedata for foreign keys to work properly.
            if connection.driver == 'sqlite':
                url = 'sqlite:///' + connection.database
                metakey = url
            else:
                url = (
                    connection.driver
                    + '+' + connection.dialect
                    + '://' + connection.username
                    + ':' + connection.password
                    + '@' + connection.host
                    + ':' + str(connection.port)
                    + '/' + connection.database
                )
                metakey = (
                    connection.host
                    + ':' + str(connection.port)
                    + '/' + connection.database
                )

            # Merge new values into connection SuperDict
            if not connection.prefix: connection.prefix = ''
            connection.merge({
                'name': name,
                'metakey': metakey,
                'url': url
            })
            #connections.append(connection)
        self.package.database.connections = config
        self.package.database.connection_default = default


    def connectionsORIG(self, config: Dict, default: str):
        # Here we translate a config connectoin dictionary into an actual Connection Class
        connections = []
        for name, connection in config.items():
            # Metakey cannot be the connection name.  If 2 connections share the exact
            # same database (host, port, dbname) then they need to also share the same
            # metedata for foreign keys to work properly.
            if connection.get('driver') == 'sqlite':
                url = 'sqlite:///' + connection.get('database')
                metakey = url
            else:
                url = (connection.get('driver')
                    + '+' + connection.get('dialect')
                    + '://' + connection.get('username')
                    + ':' + connection.get('password')
                    + '@' + connection.get('host')
                    + ':' + str(connection.get('port'))
                    + '/' + connection.get('database')
                )
                metakey = (connection.get('host')
                    + ':' + str(connection.get('port'))
                    + '/' + connection.get('database')
                )

            connections.append(Connection(
                name=name,
                driver=connection.get('driver'),
                dialect=connection.get('dialect'),
                host=connection.get('host'),
                port=connection.get('port'),
                database=connection.get('database'),
                username=connection.get('username'),
                password=connection.get('password'),
                prefix=connection.get('prefix') or '',
                metakey=metakey,
                url=url,
            ))
        self._add_db_definition('connections', connections)
        self._add_db_definition('connection_default', default)


    def models(self, items: List):
        # Default registration
        self.package.registers.defaults({'models': True})

        # Register models only if allowed
        if self.package.registers.models:
            self.package.database.models = items

    def tables(self, items: List):
        # Default registration
        self.package.registers.defaults({'tables': True})

        # Register tables only if allowed
        if self.package.registers.tables:
            self.package.database.tables = items

    def seeders(self, items: List):
        # Default registration
        self.package.registers.defaults({'seeders': True})

        # Register seeders only if allowed
        if self.package.registers.seeders:
            self.package.database.seeders = items



from typing import Any


def url(context: dict, name: str, **path_params: Any) -> str:
    request = context["request"]
    return request.url_for(name, **path_params)

def asset(context: dict, path: str) -> str:
    request = context["request"]
    return request.url_for('assets', path=path)



import sqlalchemy as sa

import uvicore
from uvicore.database import Table
from uvicore.support.dumper import dd, dump


@uvicore.table()
class Users(Table):

    # Actual database table name
    # Plural table names and singluar model names are encouraged
    # Do not add a package prefix, leave that to the connection config
    name = 'users'

    # Connection for this database from your config file
    connection = 'auth'

    # SQLAlchemy Table definition as a list (exclude name and metadata)
    # This will be converted into an actual SQLAlchemy Table() instance
    # See https://docs.sqlalchemy.org/en/13/core/schema.html
    schema = [
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("email", sa.String(length=50))
    ]

    # Optional SQLAlchemy Table() instance kwargs
    schema_kwargs = {}

#Users = Users()
# IoC Class Instance
#Users = _Users
#Users: _Users = uvicore.ioc.make('uvicore.auth.database.tables.users.Users', _Users, singleton=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uvicore',
 'uvicore.auth',
 'uvicore.auth.commands',
 'uvicore.auth.config',
 'uvicore.auth.contractsOLD',
 'uvicore.auth.database.seeders',
 'uvicore.auth.database.tables',
 'uvicore.auth.models',
 'uvicore.configuration',
 'uvicore.configuration.commands',
 'uvicore.console',
 'uvicore.console.commands',
 'uvicore.console.commands.stubs',
 'uvicore.container',
 'uvicore.container.commands',
 'uvicore.contracts',
 'uvicore.database',
 'uvicore.database.OBSOLETE',
 'uvicore.database.commands',
 'uvicore.database.commands.stubs',
 'uvicore.events',
 'uvicore.events.commands',
 'uvicore.factories',
 'uvicore.foundation',
 'uvicore.foundation.config',
 'uvicore.foundation.decorators',
 'uvicore.http',
 'uvicore.http.OBSOLETE',
 'uvicore.http.OBSOLETE.controllers',
 'uvicore.http.commands',
 'uvicore.http.routing',
 'uvicore.http.templating',
 'uvicore.logging',
 'uvicore.orm',
 'uvicore.orm.commands',
 'uvicore.orm.commands.stubs',
 'uvicore.package',
 'uvicore.package.commands',
 'uvicore.support',
 'uvicore.typing']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'aiofiles>=0.6.0,<0.7.0',
 'alembic>=1.5.2,<2.0.0',
 'asyncclick>=7.1.2,<8.0.0',
 'colored>=1.4.2,<2.0.0',
 'databases>=0.4.1,<0.5.0',
 'environs>=9.3.0,<10.0.0',
 'fastapi>=0.63.0,<0.64.0',
 'gunicorn>=20.0.4,<21.0.0',
 'ipython>=7.19.0,<8.0.0',
 'prettyprinter>=0.18.0,<0.19.0',
 'requests>=2.25.1,<3.0.0',
 'uvicorn>=0.13.3,<0.14.0']

setup_kwargs = {
    'name': 'uvicore',
    'version': '0.1.5',
    'description': '',
    'long_description': None,
    'author': 'Matthew Reschke',
    'author_email': 'mail@mreschke.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from uvicore.support.dumper import dump, dd
from uvicore.support.module import location
from uvicore.typing import List, Dict

class Http:
    """Http Service Provider Mixin"""

    def _add_http_definition(self, key, value):
        if 'http' not in self.package:
            self.package['http'] = Dict()
        self.package['http'][key] = value

    def web_routes(self, module: str, prefix: str):
        # Default registration
        self.package.registers.defaults({'web_routes': True})

        # Register WEB routes only if allowed
        if self.package.registers.web_routes:
            self._add_http_definition('web_routes', module)
            self._add_http_definition('web_route_prefix', prefix)

    def api_routes(self, module: str, prefix: str):
        # Default registration
        self.package.registers.defaults({'api_routes': True})

        # Register API routes only if allowed
        if self.package.registers.api_routes:
            self._add_http_definition('api_routes', module)
            self._add_http_definition('api_route_prefix', prefix)

    def middleware(self, items: List):
        # Default registration
        self.package.registers.defaults({'middleware': True})

        # Register middleware only if allowed
        if self.package.registers.middleware:
            self._add_http_definition('middleware', items)

    def views(self, items: List):
        # Default registration
        self.package.registers.defaults({'views': True})

        # Register views only if allowed
        if self.package.registers.views:
            self._add_http_definition('view_paths', items)

    def assets(self, items: List):
        # Default registration
        self.package.registers.defaults({'assets': True})

        # Register assets only if allowed
        if self.package.registers.assets:
            self._add_http_definition('asset_paths', items)

    def template(self, items: Dict):
        # Default registration - template obeys view registration
        self.package.registers.defaults({'views': True})

        # Register templates only if [views] allowed
        if self.package.registers.views:
            self._add_http_definition('template_options', Dict(items))



import uvicore
from uvicore.support.dumper import dump, dd
from dataclasses import dataclass
from typing import Dict, Optional

#@dataclass
@uvicore.service()
class Event:

    def dispatch(self):
        uvicore.events._dispatch(self)

    # No, logging should be a listener
    # def log(self, cls):
    #     uvicore.log.debug("Event " + str(cls.__class__.__module__) + " Dispatched")



#from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Dict, Generic, TypeVar, Union, List, Tuple, Any

class Mapper(ABC):

    @abstractmethod
    def column(self) -> str:
        """Convert a model field name into a table column name"""
        pass



import uvicore
from uvicore.typing import Dict, Any
from uvicore.contracts import Connection
from uvicore.support.dumper import dump, dd
from prettyprinter import pretty_call, register_pretty
from uvicore.contracts import Package as PackageInterface


@uvicore.service(aliases=['Package', 'package'])
class Package(PackageInterface):

    @property
    def config(self) -> Dict:
        """Helper to access this packages configs"""
        return uvicore.config.dotget(self.name)

    # def config(self, dotkey: str = None) -> Any:
    #     """Helper to access this packages configs"""
    #     if dotkey:
    #         return uvicore.config(self.name + '.' + dotkey)
    #     else:
    #         return uvicore.config(self.name)

    def connection(self, name: str = None) -> Connection:
        """Helper to access this packages connections"""
        try:
            if not self.database.connections: return None
            if name:
                return next(connection for connection in self.database.connections if connection.name == name)
            else:
                return next(connection for connection in self.database.connections if connection.name == self.database.connection_default)
        except:
            return None


@register_pretty(Package)
def pretty_entity(value, ctx):
    """Custom pretty printer for my SuperDict"""
    # This printer removes the class name uvicore.types.Dict and makes it print
    # with a regular {.  This really cleans up the output!

    # SuperDict are printed as Dict, but this Package SuperDict should
    # be printed more like a class with key=value notation, so use **values
    return pretty_call(ctx, 'Package', **value)




# # OBSOLETE
# @uvicore.service(aliases=['Package', 'package'])
# class Package(PackageInterface):

#     def config(self, dotkey: str = None):
#         if dotkey:
#             return uvicore.config(self.name + '.' + dotkey)
#         else:
#             return uvicore.config(self.name)

#     def connection(self, name: str = None):
#         if name:
#             return next(connection for connection in self.connections if connection.name == name)
#         else:
#             #return next(connection for connection in self.connections if connection.default == True)
#             if 'database' in self.config():
#                 return next(connection for connection in self.connections if connection.name == self.config('database.default'))



# IoC Class Instance
# No because not to be used by the public
#Package: _Package = uvicore.ioc.make('Package', _Package, aliases=['package'])

# Public API for import * and doc gens
#__all__ = ['_Package']



from __future__ import annotations

import operator as operators
from copy import copy
from typing import Any, Dict, Generic, List, Tuple, TypeVar, Union

import sqlalchemy as sa
from sqlalchemy.sql.expression import BinaryExpression
from sqlalchemy.engine.result import RowProxy

import uvicore
from uvicore.database.builder import _QueryBuilder, Join
from uvicore.support.dumper import dd, dump
from uvicore.contracts import DbQueryBuilder as BuilderInterface

B = TypeVar("B")  # Builder Type (DbQueryBuilder or OrmQueryBuilder)
E = TypeVar("E")  # Entity Model

@uvicore.service()
class _DbQueryBuilder(Generic[B, E], _QueryBuilder[B, E], BuilderInterface[B, E]):
    """Database Query Builder"""

    def __init__(self, connection: str):
        self._conn = connection
        super().__init__()

    def table(self, table: Union[str, sa.Table]) -> B[B, E]:
        if type(table) == str:
            self.query.table = uvicore.db.table(table, self._connection())
        else:
            self.query.table = table
        return self

    def select(self, *args) -> B[B, E]:
        for column in args:
            self.query.selects.append(column)
        return self

    def join(self, table: Union[str, sa.Table], left_where: Union[str, sa.Column, BinaryExpression], right_where: Union[str, sa.Column] = None, alias: str = None, method: str = 'join') -> B[B, E]:
        # Get table and tablename
        conn = self._connection()
        if type(table) == str:
            if '.' in table: conn, table = tuple(table.split('.'))
            table = uvicore.db.table(table, conn)
        tablename = str(table.name)

        # Get left, right and onclause expressions
        left = None
        right = None
        if type(left_where) == BinaryExpression:
            onclause = left_where
        else:
            left = self._column(left_where)
            right = self._column(right_where)
            onclause = left.sacol == right.sacol

        # Set alias to tablename if not defined
        if not alias: alias = tablename

        # Add new Join() expression
        self.query.joins.append(Join(table, tablename, left, right, onclause, alias, method))
        return self

    def outer_join(self, table: Union[str, sa.Table], left_where: Union[str, sa.Column, BinaryExpression], right_where: Union[str, sa.Column] = None, alias: str = None) -> B[B, E]:
        self.join(table=table, left_where=left_where, right_where=right_where, method='outerjoin', alias=alias)
        return self

    def group_by(self, *args) -> B[B, E]:
        for group_by in args:
            self.query.group_by.append(group_by)
        return self

    async def find(self, pk_value: Any) -> RowProxy:
        # Where on Primary Key
        self.where(self._pk(), pk_value)

        # Build query
        query, saquery = self._build_query('select', copy(self.query))

        # Execute query
        results = await uvicore.db.fetchone(saquery, connection=self._connection())

        return results

    async def get(self) -> List[RowProxy]:
        # Build query
        query, saquery = self._build_query('select', copy(self.query))

        # Execute query
        results = await uvicore.db.fetchall(saquery, connection=self._connection())

        return results

# IoC Class Instance
#_DbQueryBuilderIoc: _DbQueryBuilder = uvicore.ioc.make('DbQueryBuilder', _DbQueryBuilder)

# Actual Usable Model Class Derived from IoC Inheritence
#class DbQueryBuilder(Generic[B, E], _DbQueryBuilderIoc[B, E], BuilderInterface[B, E]):
    #pass
